"""
    Simple script to test Weave server support and ensure that it
    works properly.
"""

import sys
import urllib
import urllib2
import httplib
from urlparse import urlsplit
from xml.etree import cElementTree as ET

import json
import weave_server
import threading

class DavRequest(urllib2.Request):
    def __init__(self, method, *args, **kwargs):
        urllib2.Request.__init__(self, *args, **kwargs)
        self.__method = method

    def get_method(self):
        return self.__method

class DavHandler(urllib2.BaseHandler):
    def _normal_response(self, req, fp, code, msg, headers):
        return fp

    # Multi-status
    http_error_207 = _normal_response

    # Created
    http_error_201 = _normal_response

    # Accepted
    http_error_202 = _normal_response

    # No content
    http_error_204 = _normal_response

class WeaveSession(object):
    def __init__(self, username, password, server_url,
                 realm=weave_server.DEFAULT_REALM):
        self.username = username
        self.server_url = server_url
        self.server = urlsplit(server_url).netloc
        self.realm = realm
        self.password = password

    def clone(self):
        return WeaveSession(self.username, self.password,
                            self.server_url, self.realm)

    def _open(self, req):
        davHandler = DavHandler()
        authHandler = urllib2.HTTPBasicAuthHandler()
        authHandler.add_password(self.realm,
                                 self.server,
                                 self.username,
                                 self.password)
        opener = urllib2.build_opener(authHandler, davHandler)
        if isinstance(req, urllib2.Request):
            print req.get_data()
            print req.get_full_url()
        else:
            print req
        return opener.open(req)

    def _get_user_url(self, path, user=None):
        if not user:
            user = self.username
        if path.startswith("/"):
            path = path[1:]
        url = "%s/user/%s/%s" % (self.server_url,
                                 user,
                                 path)
        return url

    def list_files(self, path):
        xml_data = ("<?xml version=\"1.0\" encoding=\"utf-8\" ?>"
                    "<D:propfind xmlns:D='DAV:'><D:prop/></D:propfind>")

        url = self._get_user_url(path)
        headers = {"Content-type" : "text/xml; charset=\"utf-8\"",
                   "Depth" : "1"}
        req = DavRequest("PROPFIND", url, xml_data, headers=headers)
        result_xml = self._open(req).read()

        multistatus = ET.XML(result_xml)
        hrefs = multistatus.findall(".//{DAV:}href")
        root = hrefs[0].text
        return [href.text[len(root):] for href in hrefs[1:]]

    def create_dir(self, path):
        req = DavRequest("MKCOL", self._get_user_url(path))
        self._open(req)

    def remove_dir(self, path):
        if not path[-1] == "/":
            path += "/"
        self.delete_file(path)

    def get_file(self, path, user=None):
        obj = self._open(self._get_user_url(path, user))
        return obj.read()

    def put_file(self, path, data):
        req = DavRequest("PUT", self._get_user_url(path), data)
        self._open(req)

    def delete_file(self, path):
        req = DavRequest("DELETE", self._get_user_url(path))
        self._open(req)

    def lock_file(self, path):
        headers = {"Content-type" : "text/xml; charset=\"utf-8\"",
                   "Depth" : "infinity",
                   "Timeout": "Second-600"}
        xml_data = ("<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n"
                    "<D:lockinfo xmlns:D=\"DAV:\">\n"
                    "  <D:locktype><D:write/></D:locktype>\n"
                    "  <D:lockscope><D:exclusive/></D:lockscope>\n"
                    "</D:lockinfo>")
        req = DavRequest("LOCK", self._get_user_url(path), xml_data,
                         headers=headers)
        result_xml = self._open(req).read()

        response = ET.XML(result_xml)
        token = response.find(".//{DAV:}locktoken/{DAV:}href").text
        return token

    def unlock_file(self, path, token):
        headers = {"Lock-Token" : "<%s>" % token}
        req = DavRequest("UNLOCK", self._get_user_url(path),
                         headers=headers)
        self._open(req)

    def ensure_unlock_file(self, path):
        xml_data = ("<?xml version=\"1.0\" encoding=\"utf-8\" ?>"
                    "<D:propfind xmlns:D='DAV:'>"
                    "<D:prop><D:lockdiscovery/></D:prop></D:propfind>")

        url = self._get_user_url(path)
        headers = {"Content-type" : "text/xml; charset=\"utf-8\"",
                   "Depth" : "0"}
        req = DavRequest("PROPFIND", url, xml_data, headers=headers)
        try:
            result_xml = self._open(req).read()
        except urllib2.HTTPError, e:
            return

        multistatus = ET.XML(result_xml)
        href = multistatus.find(".//{DAV:}locktoken/{DAV:}href")
        if href is not None:
            self.unlock_file(path, href.text)

    def does_email_exist(self, email):
        return self._does_entity_exist("chkmail", email)

    def does_username_exist(self, username):
        return self._does_entity_exist("check", username)

    def _does_entity_exist(self, entity_kind, entity):
        url = "%s/api/register/%s/%s" % (self.server_url,
                                         entity_kind,
                                         entity)
        result = int(self._open(url).read())
        print "result: %r" % result
        if result == 0:
            return True
        elif result == 1:
            return False
        else:
            raise Exception("Unexpected result code: %d" % result)

    def change_password(self, new_password):
        url = "%s/api/register/chpwd/" % (self.server_url)
        postdata = urllib.urlencode({"uid" : self.username,
                                     "password" : self.password,
                                     "new" : new_password})
        req = urllib2.Request(url, postdata)
        print "change password reponse: %r" % self._open(req).read()
        self.password = new_password

    def share_with_users(self, path, users):
        url = "%s/api/share/" % (self.server_url)
        cmd = {"version" : 1,
               "directory" : path,
               "share_to_users" : users}
        postdata = urllib.urlencode({"cmd" : json.write(cmd),
                                     "uid" : self.username,
                                     "password" : self.password})
        req = urllib2.Request(url, postdata)
        result = self._open(req).read()
        if result != "OK":
            raise Exception("Share attempt failed: %s" % result)

def ensure_weave_disallows_php(session):
    print "Ensuring that weave disallows PHP upload and execution."
    session.put_file("phptest.php", "<?php echo 'hai2u!' ?>")
    try:
        if session.get_file("phptest.php") == "hai2u!":
            raise Exception("Weave server allows PHP execution!")
    finally:
        session.delete_file("phptest.php")

def _do_test(session_1, session_2):
    print "Ensuring that user '%s' exists." % session_1.username
    assert session_1.does_username_exist(session_1.username)

    print "Ensuring that user '%s' exists." % session_2.username
    assert session_1.does_username_exist(session_2.username)

    print "Changing password of user '%s' to 'blarg'." % session_1.username
    old_pwd = session_1.password
    old_session = session_1.clone()
    try:
        session_1.change_password("blarg")
    except urllib2.HTTPError, e:
        if (e.code == httplib.BAD_REQUEST and
            e.read() == weave_server.WeaveApp.ERR_INCORRECT_PASSWORD):
            print ("That didn't work; an old run of this test may "
                   "have been aborted.  Trying to revert...")
            session_1.password = "blarg"
            session_1.change_password(old_pwd)
            print "Revert successful, attempting to change password again."
            session_1.change_password("blarg")
        else:
            raise

    try:
        print "Ensuring we can't log in using old password."
        old_session.change_password("fnarg")
    except urllib2.HTTPError, e:
        if e.code != httplib.BAD_REQUEST:
            raise
        content = e.read()
        print "Content: %r" % content
        if content != weave_server.WeaveApp.ERR_INCORRECT_PASSWORD:
            raise AssertionError("Bad return value: %s" % content)
    else:
        raise AssertionError("We could log in using the old password!")

    print "Reverting back to old password."
    session_1.change_password(old_pwd)

    print
    print "*" * 79
    print

    print "Ensuring that file is not locked."
    session_1.ensure_unlock_file("test_lock")

    print "Locking file"
    session_1.lock_file("test_lock")

    print "Unlocking file by querying for its token"
    session_1.ensure_unlock_file("test_lock")

    print "Locking file again"
    token = session_1.lock_file("test_lock")

    try:
        print "Ensuring that we can't re-lock the file."
        session_1.lock_file("test_lock")
    except urllib2.HTTPError, e:
        if e.code != httplib.LOCKED:
            raise
    else:
        raise AssertionError("We can re-lock the file!")

    print "Unlocking file"
    session_1.unlock_file("test_lock", token)

    # FIXME
#    print "Ensuring that PROPFIND on the user's home dir works."
#    files = session_1.list_files("")

#    print "Cleaning up any files left over from a failed previous test."
#    if "blargle/bloop" in files:
#        session_1.delete_file("blargle/bloop")
#    if "blargle/" in files:
#        session_1.remove_dir("blargle")

    print "Creating directory."
    session_1.create_dir("blargle")

    print "Ensuring that directory indexes don't raise errors."
    session_1.get_file("")

    try:
        print "Creating temporary file."
        session_1.put_file("blargle/bloop", "hai2u!")
        print "Verifying that temporary file is listed."
        assert "bloop" in session_1.list_files("blargle/")
        try:
            assert session_1.get_file("blargle/bloop") == "hai2u!"
            session_1.share_with_users("blargle", [])
            try:
                print "Ensuring user 2 can't read user 1's file."
                session_2.get_file("blargle/bloop", session_1.username)
            except urllib2.HTTPError, e:
                if e.code != httplib.UNAUTHORIZED:
                    raise
            else:
                raise AssertionError("User 2 can read user 1's file!")
            print "Sharing directory with user 2."
            session_1.share_with_users("blargle", [session_2.username])
            print "Ensuring user 2 can read user 1's file."
            assert session_2.get_file("blargle/bloop",
                                      session_1.username) == "hai2u!"
            print "Sharing directory with everyone."
            session_1.share_with_users("blargle", ["all"])
            print "Ensuring user 2 can read user 1's file."
            assert session_2.get_file("blargle/bloop",
                                      session_1.username) == "hai2u!"
        finally:
            session_1.delete_file("blargle/bloop")
    finally:
        print "Removing directory."
        session_1.remove_dir("blargle")

    ensure_weave_disallows_php(session_1)

    print "Test complete."

def redirect_stdio(func):
    def wrapper(*args, **kwargs):
        from cStringIO import StringIO
        old_stdio = [sys.stdout, sys.stderr]
        stream = StringIO()
        sys.stderr = sys.stdout = stream
        try:
            try:
                return func(*args, **kwargs)
            except Exception, e:
                import traceback
                traceback.print_exc()
                raise Exception("Test failed:\n\n%s" % stream.getvalue())
        finally:
            sys.stderr, sys.stdout = old_stdio

    wrapper.__name__ = func.__name__
    return wrapper

@redirect_stdio
def test_weave_server():
    server_url = "http://127.0.0.1:%d" % weave_server.DEFAULT_PORT
    username_1 = "foo"
    password_1 = "test123"
    username_2 = "bar"
    password_2 = "test1234"

    start_event = threading.Event()

    def run_server():
        app = weave_server.WeaveApp()
        app.add_user(username_1, password_1)
        app.add_user(username_2, password_2)
        httpd = weave_server.make_server('', weave_server.DEFAULT_PORT, app)
        start_event.set()
        while 1:
            request, client_address = httpd.socket.accept()
            httpd.process_request(request, client_address)

    thread = threading.Thread(target=run_server)
    thread.setDaemon(True)
    thread.start()

    start_event.wait()

    session_1 = WeaveSession(username_1, password_1, server_url)
    session_2 = WeaveSession(username_2, password_2, server_url)

    _do_test(session_1, session_2)

if __name__ == "__main__":
    args = sys.argv[1:]
    if len(args) < 5:
        print ("usage: %s <server-url> <username-1> <password-1> "
               "<username-2> <password-2>" % sys.argv[0])
        sys.exit(1)

    server_url = args[0]
    username_1 = args[1]
    password_1 = args[2]
    username_2 = args[3]
    password_2 = args[4]
    session_1 = WeaveSession(username_1, password_1, server_url)
    session_2 = WeaveSession(username_2, password_2, server_url)

    _do_test(session_1, session_2)



'''
Utility functions for Weave API.

Created on 15.03.2010

@license: GNU GPL v3 or above, see LICENSE for more details.
@copyright: 2010 see AUTHORS for more details.
@author: Jens Diemer
@author: FladischerMichael
'''

import time
from datetime import datetime
    
def weave_timestamp(timedata=None):
    if timedata is None:
        timedata = datetime.now()
    return time.mktime(timedata.timetuple())

def limit_wbo_queryset(request, queryset):
    """
    TODO:
    predecessorid = fromform(form, "predecessorid")
    full = fromform(form, "full")
    """
    GET = request.GET

    ids = GET.get("ids", None)
    if ids is not None:
        ids = ids.split(",")
        queryset = queryset.filter(wboid__in=ids)

    parentid = GET.get("parentid", None)
    if parentid is not None:
        queryset = queryset.filter(parentid=parentid)

    newer = GET.get("newer", None)
    if newer is not None: # Greater than or equal to newer modified timestamp
        queryset = queryset.filter(modified__gte=datetime.fromtimestamp(float(newer)))

    older = GET.get("older", None)
    if older is not None: # Less than or equal to older modified timestamp
        queryset = queryset.filter(modified__lte=datetime.fromtimestamp(float(older)))

    index_above = GET.get("index_above", None)
    if index_above is not None: # Greater than or equal to index_above modified timestamp
        queryset = queryset.filter(sortindex__gte=int(index_above))

    index_below = GET.get("index_below", None)
    if index_below is not None: # Less than or equal to index_below modified timestamp
        queryset = queryset.filter(sortindex__lte=int(index_below))

    sort_type = GET.get("sort", None)
    if sort_type is not None:
        if sort_type == 'oldest':
            queryset = queryset.order_by("modified")
        elif sort_type == 'newest':
            queryset = queryset.order_by("-modified")
        elif sort_type == 'index':
            queryset = queryset.order_by("wboid")
        else:
            raise NameError("sort type %r unknown" % sort_type)

    offset = GET.get("offset", None)
    if offset is not None:
        queryset = queryset[int(offset):]

    limit = GET.get("limit", None)
    if limit is not None:
        queryset = queryset[:int(limit)]

    return queryset
        


# coding: utf-8
"""
    testproject.settings
    ~~~~~~~~~~~~~~~~~~~~~~

    IMPORTANT:
        You should not edit this file!
        Owerwrite settings with a local settings file:
            local_settings.py

    Here are not all settings predifined you can use. Please look at the
    django documentation for a full list of all items:
        http://www.djangoproject.com/documentation/settings/

    Last commit info:
    ~~~~~~~~~~~~~~~~~
    $LastChangedDate:$
    $Rev:$
    $Author: JensDiemer $

    :copyleft: 2010 by the django-weave team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import os
import sys

try:
    #from django_tools.utils import info_print;info_print.redirect_stdout()
    import django
    import weave
except Exception, e:
    import traceback
    print "-" * 79
    sys.stderr.write(traceback.format_exc())
    print "-" * 79
    raise

PROJECT_ROOT = os.path.abspath(os.path.dirname(__file__))

DEBUG = True
TEMPLATE_DEBUG = DEBUG

ADMINS = () 

MANAGERS = ADMINS

DATABASE_ENGINE = 'sqlite3'
DATABASE_NAME = os.path.join(PROJECT_ROOT, 'test.db3')

TIME_ZONE = "UTC"

LANGUAGE_CODE = 'en-us'

SITE_ID = 1

USE_I18N = True

MEDIA_ROOT = ''

MEDIA_URL = ''

ADMIN_MEDIA_PREFIX = '/media/'

SECRET_KEY = "Make this unique, and don't share it with anybody!"

TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.load_template_source',
    'django.template.loaders.app_directories.load_template_source',
)

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.csrf.middleware.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.middleware.transaction.TransactionMiddleware',
)

ROOT_URLCONF = 'testproject.urls'


TEMPLATE_DIRS = (
    os.path.join(os.path.abspath(os.path.dirname(django.__file__)), "contrib/admin/templates"),
)

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.admin',
    'django.contrib.admindocs',
    "weave",
)

# Must be obtained from http://recaptcha.net/ by registering an account.
RECAPTCHA_PUBLIC_KEY = '' 
RECAPTCHA_PRIVATE_KEY = ''


'''
URL mapping.

Created on 15.03.2010

@license: GNU GPL v3 or above, see LICENSE for more details.
@copyright: 2010 see AUTHORS for more details.
@author: Jens Diemer
@author: FladischerMichael
'''

from django.conf.urls.defaults import patterns

from weave.views import sync, user, misc

urlpatterns = patterns('weave',
    (r'^(?P<version>[\d\.]+?)/(?P<username>.*?)/storage/(?P<col_name>.*?)/(?P<wboid>.*?)$', sync.storage),
    (r'^(?P<version>[\d\.]+?)/(?P<username>.*?)/storage/(?P<col_name>.*?)$', sync.storage),
    (r'^(?P<version>[\d\.]+?)/(?P<username>.*?)/storage$', sync.storage),
    (r'^(?P<version>[\d\.]+?)/(?P<username>.*?)/info/collections$', sync.info),
    (r'^misc/(?P<version>[\d\.]+?)/captcha_html$', misc.captcha),
    (r'^user/(?P<version>[\d\.]+?)/(?P<username>.*?)/node/weave$', user.node),
    (r'^user/(?P<version>[\d\.]+?)/(?P<username>.*?)/password$', user.password),
    (r'^user/(?P<version>[\d\.]+?)/(?P<username>.*?)$', user.exists),
    (r'^api/register/check/(?P<username>.*?)$', user.register_check),
)



# coding: utf-8

from django.conf.urls.defaults import patterns, include, url

from django.contrib import admin
admin.autodiscover()

handler404 = 'django.views.defaults.page_not_found'
handler500 = 'django.views.defaults.server_error'

urlpatterns = patterns('',
    url(r'^weave/', include('weave.urls')),

    url(r'^admin/doc/', include('django.contrib.admindocs.urls')),
    url(r'^admin/', include(admin.site.urls)),
)



"""
This file demonstrates two different styles of tests (one doctest and one
unittest). These will both pass when you run "manage.py test".

Replace these with more appropriate tests for your application.

    :copyleft: 2010 by the django-weave team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from django.test import TestCase

class SimpleTest(TestCase):
    def test_basic_addition(self):
        """
        Tests that 1 + 1 always equals 2.
        """
        self.failUnlessEqual(1 + 1, 2)

__test__ = {"doctest": """
Another way to test that 1 + 1 is equal to 2.

>>> 1 + 1 == 2
True
"""}




'''
from http://hg.mozilla.org/labs/weave/file/tip/tools/scripts/weave_server.py#l189

Created on 15.03.2010

@license: GNU GPL v3 or above, see LICENSE for more details.
@copyright: 2010 see AUTHORS for more details.
@author: Jens Diemer
@author: Michael Fladischer <michael@fladi.at>
'''

ERR_UID_OR_EMAIL_AVAILABLE = "1"
ERR_WRONG_HTTP_METHOD = "-1"
ERR_MISSING_UID = "-2"
ERR_INVALID_UID = "-3"
ERR_UID_OR_EMAIL_IN_USE = "0"
ERR_EMAIL_IN_USE = "-5"
ERR_MISSING_PASSWORD = "-8"
ERR_MISSING_RECAPTCHA_CHALLENGE_FIELD = "-6"
ERR_MISSING_RECAPTCHA_RESPONSE_FIELD = "-7"
ERR_MISSING_NEW = "-11"
ERR_INCORRECT_PASSWORD = "-12"
ERR_ACCOUNT_CREATED_VERIFICATION_SENT = "2"
ERR_ACCOUNT_CREATED = "3"


# coding:utf-8


import os
import logging


__version__ = (0, 1, 5)
__api__ = (1, 0)


try:
    from django.utils.version import get_svn_revision
except ImportError:
    pass
else:
    path = os.path.split(os.path.abspath(__file__))[0]
    svn_revision = get_svn_revision(path)
    if svn_revision != u'SVN-unknown':
        svn_revision = svn_revision.replace("-", "").lower()
        __version__ += (svn_revision,)


VERSION_STRING = '.'.join(str(part) for part in __version__)
API_STRING = '.'.join(str(integer) for integer in __api__)


class Logging(object):
    """ 
    A private class that loads and caches some global objects.
    """
    logger = None

    def get_logger(cls):
        """ 
        Initializes and returns our logger instance.
        """
        if cls.logger is None:
            class NullHandler(logging.Handler):
                def emit(self, record):
                    pass

            cls.logger = logging.getLogger('django_weave')
            cls.logger.addHandler(NullHandler())
            cls.logger.setLevel(logging.DEBUG)

        return cls.logger
    get_logger = classmethod(get_logger)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    distutils setup
    ~~~~~~~~~~~~~~~

    Last commit info:
    ~~~~~~~~~~~~~~~~~
    $LastChangedDate$
    $Rev$
    $Author$

    :copyleft: 2010 by the django-weave team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import os
import sys

from setuptools import setup, find_packages

from weave import VERSION_STRING

PACKAGE_ROOT = os.path.dirname(os.path.abspath(__file__))

def get_authors():
    authors = []
    f = file(os.path.join(PACKAGE_ROOT, "AUTHORS"), "r")
    for line in f:
        if line.startswith('*'):
            authors.append(line[1:].strip())
    f.close()
    return authors

def get_long_description():
    f = file(os.path.join(PACKAGE_ROOT, "README"), "r")
    long_description = f.read()
    f.close()
    long_description.strip()
    return long_description

setup(
    name='django-weave',
    version=VERSION_STRING,
    description='django-weave is a Django reusable application witch implements a Firefox weave server.',
    long_description=get_long_description(),
    author=get_authors(),
    maintainer="Jens Diemer",
    maintainer_email="django-weave@jensdiemer.de",
    url='http://code.google.com/p/django-weave/',
    packages=find_packages(exclude=['testproject', 'testproject.*']),
    include_package_data=True, # include package data under svn source control
    zip_safe=False,
    classifiers=[
#        "Development Status :: 1 - Planning",
#        "Development Status :: 2 - Pre-Alpha",
#        "Development Status :: 3 - Alpha",
        "Development Status :: 4 - Beta",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
#        "Intended Audience :: Education",
#        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python",
        'Framework :: Django',
        "Topic :: Database :: Front-Ends",
        "Topic :: Documentation",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Operating System :: OS Independent",
    ]
)



"""
    Check some external libs with pkg_resources.require()
    We only create warnings on VersionConflict and DistributionNotFound exceptions.
    
    See also: ./scripts/requirements/external_apps.txt
    See also: ./scripts/requirements/libs.txt
    
    Format info for pkg_resources.require():
    http://peak.telecommunity.com/DevCenter/PkgResources#requirement-objects
"""

import warnings
import pkg_resources

def check_require(requirements):
    """
    Check a package list.
    Display only warnings on VersionConflict and DistributionNotFound exceptions.
    """
    for requirement in requirements:
        try:
            pkg_resources.require(requirement)
        except pkg_resources.VersionConflict, err:
            warnings.warn("Version conflict: %s" % err)
        except pkg_resources.DistributionNotFound, err:
            warnings.warn("Distribution not found: %s" % err)

requirements = (
    # http://code.djangoproject.com/browser/django/trunk/django/__init__.py
    "django >= 1.1",

    # http://code.google.com/p/django-tools/source/browse/trunk/django_tools/__init__.py
#    "django-tools >= 0.7.0beta",

    # http://code.google.com/p/django-reversion/source/browse/trunk/src/setup.py
#    "django-reversion >= 1.1.2",
)

check_require(requirements)
